define([
        '../../Core/Cartesian2',
        '../../Core/Cartesian3',
        '../../Core/Check',
        '../../Core/Color',
        '../../Core/ColorGeometryInstanceAttribute',
        '../../Core/createGuid',
        '../../Core/defaultValue',
        '../../Core/defined',
        '../../Core/defineProperties',
        '../../Core/destroyObject',
        '../../Core/GeometryInstance',
        '../../Core/Math',
        '../../Core/Matrix3',
        '../../Core/Matrix4',
        '../../Core/Plane',
        '../../Core/PlaneGeometry',
        '../../Core/PlaneOutlineGeometry',
        '../../Core/Quaternion',
        '../../Core/WebGLConstants',
        '../../DataSources/PlaneGeometryUpdater',
        '../../Scene/PerInstanceColorAppearance',
        '../../Scene/Primitive'
    ], function(
        Cartesian2,
        Cartesian3,
        Check,
        Color,
        ColorGeometryInstanceAttribute,
        createGuid,
        defaultValue,
        defined,
        defineProperties,
        destroyObject,
        GeometryInstance,
        CesiumMath,
        Matrix3,
        Matrix4,
        Plane,
        PlaneGeometry,
        PlaneOutlineGeometry,
        Quaternion,
        WebGLConstants,
        PlaneGeometryUpdater,
        PerInstanceColorAppearance,
        Primitive) {
    'use strict';

    var outlineColor = Color.WHITE;
    var frontColor = Color.RED.withAlpha(0.2);
    var backColor = Color.CYAN.withAlpha(0.2);
    var highlightColor = Color.WHITE.withAlpha(0.2);

    /**
     * @private
     * @ionsdk
     */
    function ClippingPlanePrimitive(options) {
        options = defaultValue(options, defaultValue.EMPTY_OBJECT);
        var clippingPlane = options.clippingPlane;
        var transform = options.transform;
        var size = options.size;

        //>>includeStart('debug', pragmas.debug);
        Check.defined('options.clippingPlane', clippingPlane);
        Check.defined('options.transform', transform);
        Check.defined('options.size', size);
        //>>includeEnd('debug');

        var normalWC = Matrix4.multiplyByPointAsVector(transform, clippingPlane.normal, new Cartesian3());
        normalWC = Cartesian3.normalize(normalWC, normalWC);

        this.show = defaultValue(options.show, true);

        this._plane = new Plane(clippingPlane.normal, clippingPlane.distance);
        this._clippingPlane = clippingPlane;

        this._backPrimitiveId = {
            name : 'back',
            plane : this
        };
        this._frontPrimitiveId = {
            name : 'front',
            plane : this
        };
        this._primitiveOutlineId = createGuid();

        this._modelMatrix = Matrix4.IDENTITY.clone();

        this._originalDistance = clippingPlane.distance;
        this._transform = Matrix4.clone(transform);
        this._normalWC = normalWC;
        this._centerPositionScratch = new Cartesian3();
        this._highlighted = false;
        this._dimensions = new Cartesian2(size, size);

        this._backPrimitive = undefined;
        this._frontPrimitive = undefined;
        this._outlinePrimitive = undefined;

        this._update = true;
    }

    defineProperties(ClippingPlanePrimitive.prototype, {
        size : {
            get : function() {
                return this._dimensions.x;
            }
        },
        clippingPlane : {
            get : function() {
                return this._clippingPlane;
            }
        },
        transform : {
            get : function() {
                return this._transform;
            }
        },
        centerPosition : {
            get : function() {
                var point = this._centerPositionScratch;
                Cartesian3.clone(Cartesian3.ZERO, point);
                return Matrix4.multiplyByPoint(this._modelMatrix, point, point);
            }
        },
        frontColor : {
            get : function() {
                return this._highlighted ? highlightColor : frontColor;
            }
        },
        backColor : {
            get : function() {
                return this._highlighted ? highlightColor : backColor;
            }
        }
    });

    ClippingPlanePrimitive.prototype.highlight = function(highlight) {
        this._highlighted = highlight;

        if (defined(this._backPrimitive)) {
            var newColor = this.backColor;
            var color = this._backPrimitive.getGeometryInstanceAttributes(this._backPrimitiveId).color;
            color[0] = Color.floatToByte(newColor.red);
            color[1] = Color.floatToByte(newColor.green);
            color[2] = Color.floatToByte(newColor.blue);
            color[3] = Color.floatToByte(newColor.alpha);
            this._backPrimitive.getGeometryInstanceAttributes(this._backPrimitiveId).color = color;
            var depthFailColor = this._backPrimitive.getGeometryInstanceAttributes(this._backPrimitiveId).depthFailColor;
            depthFailColor[0] = Color.floatToByte(newColor.red);
            depthFailColor[1] = Color.floatToByte(newColor.green);
            depthFailColor[2] = Color.floatToByte(newColor.blue);
            depthFailColor[3] = Color.floatToByte(newColor.alpha);
            this._backPrimitive.getGeometryInstanceAttributes(this._backPrimitiveId).depthFailColor = depthFailColor;

            newColor = this.frontColor;
            color = this._frontPrimitive.getGeometryInstanceAttributes(this._frontPrimitiveId).color;
            color[0] = Color.floatToByte(newColor.red);
            color[1] = Color.floatToByte(newColor.green);
            color[2] = Color.floatToByte(newColor.blue);
            color[3] = Color.floatToByte(newColor.alpha);
            this._frontPrimitive.getGeometryInstanceAttributes(this._frontPrimitiveId).color = color;
            depthFailColor = this._frontPrimitive.getGeometryInstanceAttributes(this._frontPrimitiveId).depthFailColor;
            depthFailColor[0] = Color.floatToByte(newColor.red);
            depthFailColor[1] = Color.floatToByte(newColor.green);
            depthFailColor[2] = Color.floatToByte(newColor.blue);
            depthFailColor[3] = Color.floatToByte(newColor.alpha);
            this._frontPrimitive.getGeometryInstanceAttributes(this._frontPrimitiveId).depthFailColor = depthFailColor;
        }
    };

    ClippingPlanePrimitive.prototype.reset = function() {
        this._clippingPlane.distance = this._originalDistance;
        this._update = true;
    };

    ClippingPlanePrimitive.prototype.update = function(frameState) {
        if (!this.show) {
            return;
        }

        var plane = this._plane;
        var clippingPlane = this._clippingPlane;

        if (this._update || !Plane.equals(clippingPlane, plane)) {
            this._update = false;

            this._plane = Plane.clone(clippingPlane, plane);
            this._normalWC = Matrix4.multiplyByPointAsVector(this._transform, clippingPlane.normal, this._normalWC);

            var modelMatrix;
            if (Matrix4.equals(this._transform, Matrix4.IDENTITY)) {
                modelMatrix = Matrix4.clone(Matrix4.IDENTITY, this._modelMatrix);
            } else {
                modelMatrix = PlaneGeometryUpdater.createPrimitiveMatrix(this._plane, this._dimensions, this._transform, frameState.mapProjection.ellipsoid, this._modelMatrix);
            }

            this._frontPrimitive = this._frontPrimitive && this._frontPrimitive.destroy();
            this._backPrimitive = this._backPrimitive && this._backPrimitive.destroy();
            this._outlinePrimitive = this._outlinePrimitive && this._outlinePrimitive.destroy();

            this._backPrimitive = new Primitive({
                geometryInstances : new GeometryInstance({
                    geometry : PlaneGeometry.createGeometry(new PlaneGeometry()),
                    attributes : {
                        color : ColorGeometryInstanceAttribute.fromColor(this.backColor),
                        depthFailColor : ColorGeometryInstanceAttribute.fromColor(this.backColor)
                    },
                    id : this._backPrimitiveId,
                    modelMatrix : modelMatrix
                }),
                appearance : new PerInstanceColorAppearance({
                    flat : true,
                    closed : false,
                    translucent : true,
                    renderState : {
                        cull : {
                            enabled : true
                        }
                    }
                }),
                depthFailAppearance : new PerInstanceColorAppearance({
                    flat : true,
                    closed : false,
                    translucent : true,
                    renderState : {
                        cull : {
                            enabled : true
                        }
                    }
                }),
                asynchronous : false
            });
            this._frontPrimitive = new Primitive({
                geometryInstances : new GeometryInstance({
                    geometry : PlaneGeometry.createGeometry(new PlaneGeometry()),
                    attributes : {
                        color : ColorGeometryInstanceAttribute.fromColor(this.frontColor),
                        depthFailColor : ColorGeometryInstanceAttribute.fromColor(this.frontColor)
                    },
                    id : this._frontPrimitiveId,
                    modelMatrix : modelMatrix
                }),
                appearance : new PerInstanceColorAppearance({
                    flat : true,
                    closed : false,
                    translucent : true,
                    renderState : {
                        cull : {
                            enabled : true,
                            face : WebGLConstants.FRONT
                        }
                    }
                }),
                depthFailAppearance : new PerInstanceColorAppearance({
                    flat : true,
                    closed : false,
                    translucent : true,
                    renderState : {
                        cull : {
                            enabled : true,
                            face : WebGLConstants.FRONT
                        }
                    }
                }),
                asynchronous : false
            });

            this._outlinePrimitive = new Primitive({
                geometryInstances : new GeometryInstance({
                    geometry : PlaneOutlineGeometry.createGeometry(new PlaneOutlineGeometry()),
                    attributes : {
                        color : ColorGeometryInstanceAttribute.fromColor(outlineColor)
                    },
                    id : this._primitiveOutlineId,
                    modelMatrix : modelMatrix
                }),
                appearance : new PerInstanceColorAppearance({
                    flat : true,
                    translucent : false,
                    renderState : {
                        lineWidth : 1.0
                    }
                }),
                asynchronous : false
            });
        }

        this._backPrimitive.update(frameState);
        this._frontPrimitive.update(frameState);
        this._outlinePrimitive.update(frameState);
    };

    ClippingPlanePrimitive.prototype.isDestroyed = function() {
        return false;
    };

    ClippingPlanePrimitive.prototype.destroy = function() {
        if (defined(this._backPrimitive)) {
            this._backPrimitive.destroy();
        }
        if (defined(this._frontPrimitive)) {
            this._frontPrimitive.destroy();
        }
        if (defined(this._outlinePrimitive)) {
            this._outlinePrimitive.destroy();
        }

        return destroyObject(this);
    };

    return ClippingPlanePrimitive;
});

