define([
        '../Core/defaultValue',
        '../Core/defined',
        '../Core/Cartesian3',
        '../Core/Cartographic',
        '../Core/Check',
        '../Core/Ray',
        '../Scene/Cesium3DTileset',
        '../Scene/Cesium3DTileFeature',
        '../Scene/Model'
    ], function(
        defaultValue,
        defined,
        Cartesian3,
        Cartographic,
        Check,
        Ray,
        Cesium3DTileset,
        Cesium3DTileFeature,
        Model) {
    'use strict';

    var cartesianScratch = new Cartesian3();
    var rayScratch = new Ray();

    function hide(primitiveCollection) {
        var primitiveCount = primitiveCollection.length;
        var visibilityStates = new Array(primitiveCount);
        for (var i = 0; i < primitiveCount; i++) {
            var primitive = primitiveCollection.get(i);
            visibilityStates[i] = primitive.show;
            if (primitive instanceof Cesium3DTileset || primitive instanceof Model) {
                continue;
            }
            primitive.show = false;
        }
        return visibilityStates;
    }

    function restore(primitiveCollection, visibilityStates) {
        var primitiveCount = primitiveCollection.length;
        for (var i = 0; i < primitiveCount; i++) {
            var primitive = primitiveCollection.get(i);
            primitive.show = visibilityStates[i];
        }
    }

    /**
     * Computes the world position on either the terrain or tileset from a mouse position.
     *
     * @param {Scene} scene The scene
     * @ionsdk
     * @param {Cartesian2} mousePosition The mouse position
     * @param {PrimitiveCollection} primitives The primitive collection containing primitives that shouldn't be picked
     * @param {Cartesian3} result The result position
     * @returns {Cartesian3} The position in world space
     */
    function getWorldPosition(scene, mousePosition, primitives, result) {
        //>>includeStart('debug', pragmas.debug);
        Check.defined('scene', scene);
        Check.defined('mousePosition', mousePosition);
        Check.defined('primitives', primitives);
        Check.defined('result', result);
        //>>includeEnd('debug');
        var position;
        if (scene.pickPositionSupported) {
            // Hide every primitive that isn't a tileset
            var showStates = hide(primitives);

            // Don't pick default 3x3, or scene.pick may allow a mousePosition that isn't on the tileset to pickPosition.
            var pickedObject = scene.pick(mousePosition, 1, 1);
            restore(primitives, showStates);

            if (defined(pickedObject) && (pickedObject instanceof Cesium3DTileFeature || pickedObject.primitive instanceof Cesium3DTileset || pickedObject.primitive instanceof Model)) { // check to let us know if we should pick against the globe instead
                position = scene.pickPosition(mousePosition, cartesianScratch);
                restore(primitives, showStates);

                if (defined(position)) {
                    return Cartesian3.clone(position, result);
                }
            }
        }

        if (!defined(scene.globe)) {
            return;
        }

        var ray = scene.camera.getPickRay(mousePosition, rayScratch);
        position = scene.globe.pick(ray, scene, cartesianScratch);

        if (defined(position)) {
            return Cartesian3.clone(position, result);
        }
    }

    return getWorldPosition;
});

